#!/usr/bin/perl

use Nmap::Parser;
use DBI;
#use Time::gmtime;
use Time::Local;
use Config::IniFiles;
use Term::ANSIColor;
use Net::Netmask;

# Edit configuration in /etc/cancerbero/cancerbero.cfg file. Don't touch here

my $cancerbero_ver = "0.4";

my $cfg = Config::IniFiles->new( -file => '/etc/cancerbero/cancerbero.cfg' );

#[conf]
my @range = split(/,/,$cfg->val('conf', 'range')); # It will be used only the first time
my @white_list = split(/,/,$cfg->val('conf', 'white_list'));
my $white_list_text = $cfg->val('conf', 'white_list');
my $loop = $cfg->val('conf', 'loop'); # 0 runs once.  1 runs forever.
my $delay_scan = $cfg->val('conf', 'delay_scan');
my $sweep_time = $cfg->val('conf', 'sweep_time'); # Loop time between pings sweeps when runs forever in seconds.
my $max_num_scan = $cfg->val('conf', 'max_num_scan');
my $debug = $cfg->val('conf', 'debug'); # change to 1 to activate debug.
my $check_alive = $cfg->val('conf', 'check_alive');

#[database]
my $db = $cfg->val('database', 'db');
my $host_db = $cfg->val('database', 'host_db');
my $user_db = $cfg->val('database', 'user_db');
my $pass_db = $cfg->val('database', 'pass_db');

#[nmap]
my $nmap_bin = $cfg->val('nmap', 'nmap_bin');
my $nmap_sweep_opts = $cfg->val('nmap', 'nmap_sweep_opts');
my $nmap_opts = $cfg->val('nmap', 'nmap_opts');

# Vars
my $time_offset;
my @query;
my @data;
my $priority_db;
my $ping;
my $time2;
my @date_scan;
my $ip_db;
my $d_db;
my @time_db;
my $time_seg;
my $flag;
my $ip;
my $hostname = "";
my $hostname_db;
my $status;
my $run;
my $check_ip;
my @ips_to_ping;
my $ip_pending;
my $gap_pending;
my $next_sweep;
my $rest_time;
my $block_range;


# Scan object
my $np = new Nmap::Parser;

# Conect with the database
my $db = DBI->connect("DBI:mysql:database=$db;host=$host_db", "$user_db", "$pass_db", {'RaiseError' => 1});

# Preparing database stuff
my $select_configuration_values_version = $db->prepare("SELECT cancerbero_ver FROM configuration WHERE cancerbero_ver = '$cancerbero_ver'") or die "Couldn't prepare statement: " . $db->errstr;
my $insert_configuration_values_version = $db->prepare("INSERT configuration SET cancerbero_ver = ?") or die "Couldn't prepare statement: " . $db->errstr;
my $update_configuration_values = $db->prepare("UPDATE configuration SET range = ?, white_list = ?, loop = ?, delay_scan = ?, sweep_time = ?, max_num_scan = ?, debug = ?, check_alive = ?, nmap_sweep_opts = ?, nmap_opts = ? WHERE cancerbero_ver = '$cancerbero_ver'") or die "Couldn't prepare statement: " . $db->errstr;
my $select_program_values_version = $db->prepare("SELECT cancerbero_ver FROM program_values WHERE cancerbero_ver = '$cancerbero_ver'") or die "Couldn't prepare statement: " . $db->errstr;
my $insert_program_values_version = $db->prepare("INSERT program_values SET cancerbero_ver = ?") or die "Couldn't prepare statement: " . $db->errstr;
my $update_program_values_dates = $db->prepare("UPDATE program_values SET date_next = ?, date_prev = ? WHERE cancerbero_ver = '$cancerbero_ver'") or die "Couldn't prepare statement: " . $db->errstr;
my $get_ips = $db->prepare("SELECT ip FROM hosts") or die "Couldn't prepare statement: " . $db->errstr;
my $select_ip_pending = $db->prepare("SELECT ip,hostname,d_last_scan,priority,ping FROM hosts WHERE ip = ?") or die "Couldn't prepare statement: " . $db->errstr;
my $select_num_targets = $db->prepare("SELECT count(*) FROM hosts WHERE priority = '1' and DATE(FROM_UNIXTIME(d_last_scan)) < date_sub(curdate(),interval ? day)") or die "Couldn't prepare statement: " . $db->errstr;
my $select_last_scan_pending = $db->prepare("SELECT ip FROM hosts WHERE DATE(FROM_UNIXTIME(d_last_scan)) < date_sub(curdate(),interval ? day)") or die "Couldn't prepare statement: " . $db->errstr;
my $insert_pending = $db->prepare("INSERT INTO hosts (ip,hostname,priority) VALUES (?, ?, ?)") or die "Couldn't prepare statement: " . $db->errstr;
my $update_pending = $db->prepare("UPDATE hosts SET priority = ? , ping = ? WHERE ip = ?") or die "Couldn't prepare statement: " . $db->errstr;
my $update_no_dns = $db->prepare("UPDATE hosts SET comment = ? WHERE ip = ?") or die "Couldn't prepare statement: " . $db->errstr;
my $update_hostname = $db->prepare("UPDATE hosts SET hostname = ? WHERE ip = ?") or die "Couldn't prepare statement: " . $db->errstr;
my $pending = $db->prepare('SELECT ip FROM hosts WHERE priority = 1') or die "Couldn't prepare statement: " . $db->errstr;

## Functions

# Function to check if the range is already in the DB.

sub check_ranges_new {

	my @old_range = @_;
	foreach $old_block_range (@old_range) {
	
		foreach $block_range (@range)	{
		}
	}
}
	
# Function to check if the range is already in the DB. Case not, scan and insert it in the DB.

sub check_ranges {


	foreach $block_range (@range)
	{
		my $ip = 0;
		print LOG color ("green"), "Checking if $block_range IPs are already in the DB.\n", color("reset");
		my $block = new Net::Netmask ($block_range);
		$get_ips->execute();
		while (my $db_ips = $get_ips->fetchrow_array)
		{
			if ( $block->match($db_ips) )
			{
				$ip = 0;
				if ( $debug > 0) {
			 		print color ("red"), "La IP $db_ips esta en la red $block_range\n", color("reset");
				}
				next;
			} else {
				$ip = 1;
				if ( $debug > 0) {
					print color ("green"), "La IP $db_ips NO esta en la red $block_range\n", color("reset");
				}
			}
		}
		if ($ip == 1)
		{
			print LOG color ("red"), "One or more IPs from the range $block_range are not in the DB.\n", color("reset");
			print LOG color ("green"), "Sweeping & Populating the database that range...\n", color("reset");
			# Launch sweep for all the range (is the first execution of the program)
			$np->parsescan($nmap_bin, $nmap_sweep_opts, $block_range);
			# Process the sweep and return the start of the scan.
			sweeper ($np);
		} else {
			print LOG color ("green"), "Ok, $block_range is already n the DB.\n", color("reset");
		}
	}
}


# Function to avoid to scan IPs ending in 1, 0 and 255. Its avoid the scan to the white list as well including subnets.

sub not_scan_special_ips_or_nets {
	my $scan=0;
	my $input_ips=0;
	my @ips = split (/\./, $_[0]);
	
	if ( $ips[3] =~ /(^0$|^1$|255)/ ){
		$scan = 1;
	} else {
		foreach $input_ips (@white_list){
			my $block = new Net::Netmask ($input_ips);
			my $mask = $block->bits();
			if ($mask == 32 ) {    # We compare with an unique ip
				if ( $_[0] =~ /$input_ips$/ ) {
					$scan = 1;
				}
			} else {
				if ($block->match($_[0])) {  # We compare with a subnet
					$scan = 1;
				}
			}
		}
	}
	
	return $scan;
}



# Function to check if the IP was scanned less than $delay_scan (config file) days ago.

sub ip_scan_x_days_ago {

	my @ips_to_ping;
	my $counter = 0;
	my $check_ip;

	# Ask for all the scans thats needs to be scanned (older than $delay_scan).
	$select_last_scan_pending->execute($delay_scan);

	while (my @ip_last_scan = $select_last_scan_pending->fetchrow_array) {
		$check_ip = not_scan_special_ips_or_nets($ip_last_scan[0]);
                if ($check_ip == 1) {
                        next;
                } else {
			$ips_to_ping[$counter++] = $ip_last_scan[0];
		}
	}
	
	$select_last_scan_pending->finish();
	
	return @ips_to_ping;
}


# Function to sweep and include or update in the database the status of the IP.

sub sweeper {

	my $check_ip;
	# Input arguments of the function ($_[0])
	my $np = $_[0];
	
	# Picking scan info.
	my $si = $np->get_session();
	my $start_time = $si->start_time() + $time_offset;
#	my $gm = gmtime($si->start_time() + $time_offset);
#	my $year = $gm->year + 1900;
#	my $month = $gm->mon + 1;
#	my $hour = ($gm->hour)%24;
#	my $start_date = $year."-".$month."-".$gm->mday;
#	my $start_time = $hour.":".$gm->min.":".$gm->sec;
   
	if ( $debug > 0) {

        	print LOG "\n\nScan Information:\n",
        	'Executed command: '.$si->scan_args()."\n",
        	'Number of services scanned: '.$si->numservices()."\n",
        	'Start Date: '.$start_date."\n",
        	'Start Time: '.$start_time."\n",
        	'Total time: '.($si->finish_time() - $si->start_time()).' seconds'."\n",
        	'Scan Types: ',(join ' ',$si->scan_types())."\n\n\n";
	}

	# Picking host info
	for my $host_obj ($np->all_hosts()){
	
		# Grab the IP an Status of the host
		$ip = $host_obj->ipv4_addr();
		$hostname = $host_obj->hostname();
		if (!($hostname =~ /\w\..+/)) {
			$hostname = "NO DNS";
		}	
		$status = $host_obj->status();
		
		if ( $debug > 0)  {
			print LOG "IP    : $ip\n";
			print LOG "Status : $status\n\n";
		}
		
		# Look for information of this IP
		$select_ip_pending->execute($ip);
		$priority_db = 4;
		while (@data = $select_ip_pending->fetchrow_array) {
	      		$ip_db = $data[0];
			$hostname_db = $data[1];
			$d_db = $data[2];
			$priority_db = $data[3];
			$ping = $data[4];
			
			if ( $debug > 0) {
	            		print "IP: $ip_db \n";
	            		print "Fecha: $d_db \n";
	            		print "Priority: $priority_db \n";
				print "N de Pings: $ping \n";
			}
		}
		
		$select_ip_pending->finish();
	
		# Call to the function to check if is a special IP or is in the whitelist.
		# if (not_scan_special_ips($ip)) {next;}
		$check_ip = not_scan_special_ips_or_nets($ip);
		if ($check_ip == 1) {
			next;
		}
		
		# Check the priority of this IP.
		# If the host is down
		if ($status =~ /down/){
		
			if ($priority_db == 4){
				# Inserting the IP with priority 0
				$insert_pending->execute( $ip, $hostname, "0" );
			} else {
				# Increase the number of unreacheable pings done.
				if ($ping > 3000){ # This is for extract an IP with priority 2 that never answer.
					$update_pending->execute( "0", "0", $ip);
				}
				$ping++;
				if ( ($priority_db == 1) || ($priority_db == 0) ){
					$update_pending->execute( "0", $ping, $ip);
				} elsif ( $priority_db == 2 ){
					$update_pending->execute( "2", $ping, $ip);
				}
			}
			
		# If the host is up
		} elsif ($status =~ /up/){
			if ($priority_db == 4){
				# Inserting the IP with priority 1
				$insert_pending->execute( $ip, $hostname, "1" );
	      		} else {
	      	# Insert, update o mark as NO DNS the hostname in the database.
				if ($hostname eq ""){
					$update_no_dns->execute( "NO DNS", $ip );
				} elsif ($hostname_db eq ""){
					$update_hostname->execute( $hostname, $ip );
				} elsif ($hostname ne $hostname_db){
					$update_hostname->execute( $hostname, $ip );
				} 
				# Asign priority to 1 because the host is UP.
				$update_pending->execute( "1", "0", $ip);
			}
		}			
	}
	return;
}

sub nmap_scan {
			
	print LOG color ("green"), "\nLaunching the scanners...\n", color("reset");
		
	# Obtain all the IPs pending to scan.
	$pending->execute();
	

	while ($ip_pending = $pending->fetchrow_array()) {
		
		my $actual_time = time();
		
		# Checks if the sweep interval has expired, if so, stop launching scans and return the seconds until next sweep.
		if ( $actual_time >= ($next_sweep - 600) )
		{
			print LOG color ("green"), "\nStop launching scanners...\n", color("reset");
			print LOG color ("green"), "\nWaiting util scanners endding...\n", color("reset");
			$gap_pending = `ps -eo cmd | grep "sub_nmap.pl" | grep -vc grep`;
			# Not ending while, until no sub_nmaps.pl proccess are running.
			while ( $gap_pending != 0 ){
				sleep (10);
				$gap_pending = `ps -eo cmd | grep "sub_nmap.pl" | grep -vc grep`;
			}
			# Calculate rest time until next sweep and return it if is > 0.
			$actual_time = time();
			my $r_time = $next_sweep - $actual_time;
			if ($r_time >= 0)
			{
				return $r_time;
			} else {
				return 0;
			}
		}
		# Look for a gap. The max scans is in $max_num_scan
		$gap_pending = `ps -eo cmd | grep "sub_nmap.pl" | grep -vc grep`;
		while ($gap_pending >= $max_num_scan){
			sleep (10);
			$gap_pending = `ps -eo cmd | grep "sub_nmap.pl" | grep -vc grep`;
		}
		
		# Launch scan for $ip_pending
		system "sub_nmap.pl $ip_pending $time_offset &";
	}
	
	$pending->finish();
	
	# Calculate rest time until next sweep and return it if is > 0.
	my $actual_time = time();
	my $r_time = $next_sweep - $actual_time;
	print LOG "Seconds until next sweep: $r_time\n";
	if ($r_time >= 0)
	{
		return $r_time;
	} else {
		return 0;
	}
}

#### MAIN ####

# Open log file.
open(LOG, ">>/var/log/cancerbero.log");

# $time_offset will be a constant in the rest of the program
$time_offset = 0; #por ahora no hace falta

# Check versions and update config from file to DB.
$select_configuration_values_version->execute();
#my $version_conf = $select_configuration_values_version->rows;
my @version_conf = $select_configuration_values_version->fetchrow_array();
if ((scalar(@version_conf)) == 0)
{
	$insert_configuration_values_version->execute($cancerbero_ver);
}
$update_configuration_values->execute(@range, $white_list_text, $loop, $delay_scan, $sweep_time, $max_num_scan, $debug, $check_alive, $nmap_sweep_opts, $nmap_opts);

check_ranges();

while (1){

	# Create an array of IPs to scan.
	@ips_to_ping = ip_scan_x_days_ago();
	
	# Print some stats.
	my $num_scan = @ips_to_ping;
	my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst)=localtime(time);
	print LOG color ("green"), "Beginnig of the sweep...\n";
	printf LOG "%4d-%02d-%02d %02d:%02d:%02d\n",$year+1900,$mon+1,$mday,$hour,$min,$sec;
	print LOG "Num of possible pending targets IPs: ".$num_scan."\n", color("reset");
	
	# Adds the interval between sweep to the intial scan.
	my $actual_time = time();
	$next_sweep = $actual_time + $sweep_time;
	print LOG scalar "Siguiente scaneo: ".localtime($next_sweep)."\n";
	
	# Check versions and update dates scan
	$select_program_values_version->execute();
	my $version_prog = $select_program_values_version->rows;
	if ($version_prog == 0)
	{
		$insert_program_values_version->execute($cancerbero_ver);
	}
	$update_program_values_dates->execute($next_sweep, $actual_time);

	# Store next sweep time just for the gui.
	open(DATE, ">/var/run/cancerbero/date.txt");
		print DATE scalar localtime($actual_time)."\n";
		print DATE scalar localtime($next_sweep);
	close(DATE);

	# Getting the IPs to Scan.
	if ($num_scan == 0) {
	   print LOG color ("green"), "Hey! First execution of Cancerbero.\n";
	   print LOG "Populating the database with selected ranges...\n", color("reset");
	   # Launch sweep for all the range (is the first execution of the program)
		$np->parsescan($nmap_bin, $nmap_sweep_opts, @range);
		# Process the sweep and return the start of the scan.
		sweeper ($np);
	} else {
		print LOG color ("cyan"), "Searching alive IPs...\n";
		while (@ips_to_ping > 1){
			# Launch sweep for IPs to scan in groups of 5000 due a limitation of Nmap::Parser.
			$np->parsescan($nmap_bin, $nmap_sweep_opts, @ips_to_ping[0..5000]);
			# Process the sweep and return the start of the scan.
			sweeper ($np);
			# Delete the first 5000 items of the array.
			splice (@ips_to_ping,0,5000);
		} 
	} 
	
	# Print some stats.
	$select_num_targets->execute($delay_scan);
	my @num_pending = $select_num_targets->fetchrow_array;
	print LOG color ("green"), "Num of alive pending targets IPs: ".$num_pending[0]."\n", color("reset");
	$select_num_targets->finish();

	print LOG color ("green"), "\nSleeping sweeper...\n";
	
	# Sweep has finished so we can launch nmap_scan().
	$rest_time=nmap_scan($next_sweep);
	
	# Check if the program has to run forever.
	if ( $loop == 0 ){
		print LOG color ("green"), "Sweep finish.\n", color("reset");
		print LOG color ("green"), "By, have a nice day. ;)\n", color("reset");
		exit;
	}
	
	# Print when the next sweep will be.
	print LOG scalar "Next sweep will be in: $rest_time secs (".localtime($next_sweep).")\n";
	
	# Waits the time until next sweep.
	sleep($rest_time);
	
	print LOG color ("green"), "\nWaking up sweeper...\n", color("reset");
	
}

$db->disconnect;

exit;
