#!/usr/bin/perl

use Nmap::Parser;
use DBI;
#use Time::gmtime;
use strict;
use Config::IniFiles;
use Term::ANSIColor;

# Edit configuration in config.txt file. Don't touch here

my $cfg = Config::IniFiles->new( -file => '/etc/cancerbero/cancerbero.cfg' );

my $db = $cfg->val('database', 'db');
my $host_db = $cfg->val('database', 'host_db');
my $user_db = $cfg->val('database', 'user_db');
my $pass_db = $cfg->val('database', 'pass_db');
my $nmap_bin = $cfg->val('nmap', 'nmap_bin');
my $nmap_opts = $cfg->val('nmap', 'nmap_opts');
my $debug = $cfg->val('conf', 'debug');

# Scan object
my $np = new Nmap::Parser;

# IP to scan
my $input_ip = $ARGV[0];

# Function to process the output of the scan and insert or update in the database the information
sub process_xml {

	#Vars
	my @open_ports;
	my $open_ports;
	my $ip;
	my $status;
	my $ports_scanned;
	my $host;
	my $mac;
	my $mac_vendor;
	my $os_guessed;
	my $os_known;
	my $port;
	my $state;
	my $protocol;
	my $service;
	my $fullversion;
	my @lines;
	my $line;
	my $check = 0;
	my @host_id;
	my @scan_rec_id;
	my $os_obj;
	my $port_obj;
		
	# Scan object
	my $np = new Nmap::Parser;
	
	#Conect with the database
	my $db = DBI->connect("DBI:mysql:database=$db;host=$host_db", "$user_db", "$pass_db", {'RaiseError' => 1});
	
	# Preparing database stuff
	my $select_host_id = $db->prepare("SELECT host_id from hosts WHERE ip = ? ");
#	my $insert_scans = $db->prepare("INSERT INTO scans (host_id,open_ports,ports_scanned,os_guessed,d,t) VALUES (?, ?, ?, ?, ?, ?)");
	my $insert_scans = $db->prepare("INSERT INTO scans (host_id,open_ports,ports_scanned,os_guessed,d) VALUES (?, ?, ?, ?, ?)");
	my $update_pending = $db->prepare("UPDATE hosts SET d_last_scan = ? , priority = '0' , ping = '0' WHERE ip = ?");
	my $select_scan_rec_id = $db->prepare("SELECT scan_rec_id from scans WHERE host_id = ? and d = ?");
	my $insert_ports = $db->prepare("INSERT INTO ports (scan_rec_id,port,protocol,service,state,fullversion) VALUES (?, ?, ?, ?, ?, ?)");
	my $update_pending_down = $db->prepare("UPDATE hosts SET priority = '0' WHERE ip = ?");
	
	# Get the scan info
	$np = $_[0];
	my $si = $np->get_session();
	
	my $start_date = $si->start_time() + $ARGV[1];
		
#	if ( $debug > 0) {
#		print "\n\nScan Information:\n",
#		'Executed command: '.$si->args()."\n",
#		'Number of services scanned: '.$si->numservices()."\n",
#		'Start Date: '.$start_date."\n",
#		'Start Time: '.$start_time."\n",
#		'Total time: '.($si->finish_time() - $si->start_time()).' seconds'."\n",
#		'Scan Types: ',(join ' ',$si->scan_types())."\n\n\n";
#	}
		
	
	for my $host_obj ($np->all_hosts()){
	
		# Get the host scanned info
		@open_ports = $host_obj->tcp_ports('open');
		$open_ports = @open_ports;
		$ip = $host_obj->ipv4_addr();
		$status = $host_obj->status($ip);
		$ports_scanned = $host_obj->tcp_ports_count;
		$host = $host_obj->hostname();
		$mac = $host_obj->mac_addr();
		$mac_vendor = $host_obj->mac_vendor();
		$os_obj = $host_obj->os_sig();
		$os_guessed = $os_obj->name();
		$os_known = $os_obj->osfamily();
		
		# This check marks the IP as scanned. We will use it later.
		$check = 1;
		
		# Print some stats.
		print LOG color ("green"), $host_obj->ipv4_addr()." scanned.\n", color("reset");
		
		if ( $debug > 0) {
			print 'Ports Open: '.@open_ports."\n",
			'Hostname   : '.$host_obj->hostname()."\n",
			'Status     : '.$status."\n",
			'Address    : '.$host_obj->ipv4_addr()."\n",
			'MAC        : '.$host_obj->mac_addr()."\n",
			'Mac Vendor : '.$host_obj->mac_vendor()."\n",
			'OS match   : '.$os_obj->name()."\n",
			'Ports Open : '.$open_ports."\n",
			'Ports found: '.$host_obj->tcp_port_count()."\n\n";
		}
		
		# Get the host_id from $ip in host table.
		$select_host_id->execute($ip);
		@host_id = $select_host_id->fetchrow_array;
		$select_host_id->finish();
			
		# Insert the scan info into the scans table.
#		$insert_scans->execute($host_id[0], $open_ports, $ports_scanned, $os_guessed, $start_date, $start_time);
		$insert_scans->execute($host_id[0], $open_ports, $ports_scanned, $os_guessed, $start_date);
		
		# Update the last_scan data in host table.
		$update_pending->execute($start_date, $ip);
			
		# Get the scan_rec_id from scans table.
		$select_scan_rec_id->execute($host_id[0], $start_date);
		@scan_rec_id = $select_scan_rec_id->fetchrow_array;
		$select_scan_rec_id->finish();
			
		if ( $debug > 0) {
			print "Port\tState\tProtocol\tService\tAplication\tVersion\n";
		}
	
		# Process the open ports of each host.
	 	foreach $port (@open_ports) {
			
			# Create the object ports_obj
			$port_obj = $host_obj->tcp_service($port);
			
			if ( $debug > 0) {
	     		print
					$port."\t".$host_obj->tcp_port_state($port)."\t"
					.$port_obj->proto()."\t"
					.$port_obj->name()."\t"
					.$port_obj->product()."\t"
					.$port_obj->extrainfo()."\n";
			}
	
			# Get the info of the port.
	 		$state = $host_obj->tcp_port_state($port);
	 		$protocol = $port_obj->proto();
	 		$service = $port_obj->name();
	 		$fullversion = $port_obj->product()." ".$port_obj->extrainfo()." ".$port_obj->version();
	
			# Insert the info in the ports table.
				$insert_ports->execute($scan_rec_id[0], $port, $protocol, $service, $state, $fullversion);
	
		}
	
		if ( $debug > 0) {
			print "\n\n\n\n";
		}
	} 
	
	# If the IP is down change the priority
	if ($check == 0) {
		print LOG color ("magenta"), "$ARGV[0] is Down.\n", color("reset");
		$update_pending_down->execute($ARGV[0]);
	}
	
	# Disconect from the database
	$db->disconnect;
	
	# Frees memory - helpful when dealing with memory intensive scripts
 	$np->purge();
	
	return;
}



#### MAIN ####

$np->parsescan($nmap_bin, $nmap_opts, $ARGV[0]);

open(LOG, ">>/var/log/cancerbero.log");

process_xml($np);

close(LOG);


